/*
  pedometer - a tools to retreive data and manage settings from Crivit Z31192 pedometer

  Copyright (c) 2014, Laurent Mazet <mazet@softndesign.org>
*/

#define __BDB_C__

#include <malloc.h>
#include <db.h>
#include <sys/stat.h>
#include <string.h>

#include "pedometer.h"

#include "db.h"

static char *dbfilename = NULL;

static DB *dbp = NULL;

void set_dbfilename (char *name)
{
    dbfilename  = name;
}

/* open db */

int open_db (void)
{
    static int init = 1;

    if (init) {
        if (db_create (&dbp, NULL, 0)) {
            VERBOSE (ERROR, printf ("Error initializing database structure\n"));
        } else if (dbp->open (dbp, NULL, dbfilename, NULL, DB_HASH, DB_CREATE, S_IRUSR | S_IWUSR)) {
            VERBOSE (ERROR, printf ("Error opening database '%s'\n", dbfilename));
            dbp = NULL;
        }
        init = 0;
    }
    return (dbp != NULL);
}

/* close db */

void close_db ()
{
    if (dbp)
        dbp->close (dbp, 0);
    dbp = NULL;
}

/* convert db */

int convert_db (config_t *new_config)
{
    DBT key;
    DBT data;
    config_t _db_config;
    config_t *db_config = &_db_config;

    if (!open_db ())
        return 0;

    /* look for db configuration */
    memset (&key, 0, sizeof (DBT));
    memset (&data, 0, sizeof (DBT));
    char *keyname = "config";
    key.data = keyname;
    key.size = strlen (keyname) + 1;
    data.data = db_config;
    data.ulen = sizeof (config_t);
    data.flags = DB_DBT_USERMEM;
    if (dbp->get (dbp, NULL, &key, &data, 0) == DB_NOTFOUND) {
        VERBOSE (ERROR, printf ("Config not found in database\n"));
        memset (db_config, 0, sizeof (config_t));
    } else {
        VERBOSE (DEBUG, printf ("Database system: %u\n", db_config->system));
    }

    /* check db unit system */
    if (db_config->system == system_undef_e) {
        printf ("what is the database system (1 for metric, 2 for imperial)? ");
        if ((scanf("%u", &db_config->system)) && (db_config->system >= system_unknown_e)) {
            VERBOSE (WARNING, printf ("Unknown system (%u)\n", db_config->system));
            db_config->system = system_undef_e;
        }
        if (db_config->system != system_undef_e) {
            data.size = sizeof (config_t);
            if (dbp->put (dbp, NULL, &key, &data, DB_OVERWRITE_DUP)) {
                VERBOSE (WARNING, printf ("Failed inserting config into database\n"));
            } else {
                VERBOSE (DEBUG, printf ("Insert initial db config\n"));
            }
        }
    }
    VERBOSE (DEBUG, printf ("Config: %u -> %u\n", db_config->system, new_config->system));

    if (new_config->system) {
        if (new_config->system != db_config->system) {

            /* update db config */
            VERBOSE (DEBUG, printf ("Update database configuration\n"));
            data.data = new_config;
            data.size = sizeof (config_t);
            if (dbp->put (dbp, NULL, &key, &data, DB_OVERWRITE_DUP)) {
                VERBOSE (WARNING, printf ("Failed inserting config into database\n"));
            } else {

                /* convert db */
                if (db_config->system && new_config->system) {
                    VERBOSE (DEBUG, printf ("Convert database from %s to %s\n", (db_config->system == 1) ? "metric" : "imperial", (new_config->system == 1) ? "metric" : "imperial"));
                    data_t _d;
                    data_t *d = &_d;
                    memset (&key, 0, sizeof (DBT));
                    key.flags = DB_DBT_MALLOC;
                    memset (&data, 0, sizeof (DBT));
                    data.data = d;
                    data.ulen = sizeof (data_t);
                    data.flags = DB_DBT_USERMEM;
                    DBC *cursorp;
                    dbp->cursor(dbp, NULL, &cursorp, 0);
                    int nb = 0, ret;
                    while (!(ret = cursorp->get (cursorp, &key, &data, DB_NEXT))) {
                        if (strncmp (key.data, "config", 6) != 0) {
                            d->distance = convert_gait (new_config->system, db_config->system, d->distance * 10);
                            d->distance = DIVROUND (d->distance, 1000);
                            if (dbp->put (dbp, NULL, &key, &data, DB_OVERWRITE_DUP))
                                VERBOSE (WARNING, printf ("Failed updating data into database\n"));
                            nb++;
                        }
                        free (key.data);
                    }
                    VERBOSE (DEBUG, printf ("Process %d records\n", nb));
                }
            }

        } else {
            VERBOSE (WARNING, printf ("Nothing to do\n"));
        }

    } else {
        VERBOSE (WARNING, printf ("Incoherent unit systeme (db or request), nothing to do\n"));
    }

    return 1;
}

/* fill db */

int fill_db (data_t *d)
{
    DBT key;
    DBT data;

    if (!open_db ())
        return 0;

    memset (&key, 0, sizeof (key));
    memset (&data, 0, sizeof (data));
    key.data = d->date;
    key.size = strlen (d->date) + 1;
    data.data = d;
    data.size = sizeof (data_t);
    VERBOSE (DEBUG, printf ("Inserting key '%s' into db\n", d->date));
    if (dbp->put (dbp, NULL, &key, &data, DB_OVERWRITE_DUP))
        VERBOSE (WARNING, printf ("Failed inserting into database\n"));

    return 1;
}

/* extract db */

int extract_db (data_t **table)
{
    data_t _d;
    data_t *d = &_d;
    int i, nb, ret;
    DBT key;
    DBT data;
    DBC *cursorp;

    if (!open_db ())
        return 0;

    memset (&key, 0, sizeof (DBT));
    key.flags = DB_DBT_MALLOC;
    memset (&data, 0, sizeof (DBT));
    data.data = d;
    data.ulen = sizeof (data_t);
    data.flags = DB_DBT_USERMEM;
    dbp->cursor(dbp, NULL, &cursorp, 0);

    nb = 0;
    while (!(ret = cursorp->get (cursorp, &key, &data, DB_NEXT))) {
        if (strncmp (key.data, "config", 6) != 0) {
            *table = (data_t *) realloc (*table, (nb + 2) * sizeof (data_t));
            for (i = 0; i < nb; i++)
                if (strcmp (d->date, (*table + i)->date) > 0)
                    break;
            if (i < nb)
                memmove (*table + i + 1, *table + i, (nb - i) * sizeof (data_t));
            memcpy (*table + i, d, sizeof (data_t));
            nb++;
        }
        free (key.data);
    }

    if (ret != DB_NOTFOUND)
        VERBOSE (ERROR, printf ("Nothing found in database\n"));
    if (cursorp != NULL)
        cursorp->close (cursorp);

    return nb;
}
