/*
  pedometer - a tools to retreive data and manage settings from Crivit Z31192 pedometer

  Copyright (c) 2014, Laurent Mazet <mazet@softndesign.org>
*/

/*
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* depend: */
/* cflags: -Wno-missing-field-initializers */
/* linker: -ldb -lleveldb -lmadb */

#include <db.h>
#include <getopt.h>
#include <leveldb/c.h>
#include <malloc.h>
#include <snd/madb.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

/* version */

#define VERSION "0.1"

#include "version.h"

/* global variables */

char *progname;

DB *bdbp = NULL;
leveldb_t *ldbp = NULL;
char *errptr = NULL;
madb_t *mdbp = NULL;

#define NONE 0
#define BDB 1
#define LDB 2
#define MDB 3

#define BDB2LDB ((BDB) | (LDB << 2))
#define BDB2MDB ((BDB) | (MDB << 2))
#define LDB2BDB ((LDB) | (BDB << 2))
#define LDB2MDB ((LDB) | (MDB << 2))
#define MDB2BDB ((MDB) | (BDB << 2))
#define MDB2LDB ((MDB) | (LDB << 2))

#define IS_DUMP(x) (((x) >> 2) == NONE)

int way = BDB2LDB;

/* verbose macros */

int verbose = 1;

#define ERROR 0
#define WARNING 1
#define DEBUG 2

#define VERBOSE(level, statement...)\
    do {                                                         \
        if (level <= verbose) {                                  \
            printf ("\033[1;%dm", (level == ERROR) ? 31 :        \
                                  (level == WARNING) ? 33 : 34); \
            statement;                                           \
            printf ("\033[0;0m");                                \
        }                                                        \
    } while(0)

/* open database */

int open_dbs (char *indbname, char *outdbname, int way)
{
    char *bdbname = NULL;
    char *ldbname = NULL;
    char *mdbname = NULL;

    switch (way) {
    case BDB:
    case BDB2LDB:
    case BDB2MDB:
        bdbname = indbname;
        break;
    case LDB:
    case LDB2BDB:
    case LDB2MDB:
        ldbname = indbname;
        break;
    case MDB:
    case MDB2BDB:
    case MDB2LDB:
        mdbname = indbname;
        break;
    default:
        VERBOSE (ERROR, printf ("Unknown conversion way\n"));
        return 0;
    }

    switch (way) {
    case LDB2BDB:
    case MDB2BDB:
        bdbname = outdbname;
        break;
    case BDB2LDB:
    case MDB2LDB:
        ldbname = outdbname;
        break;
    case BDB2MDB:
    case LDB2MDB:
        mdbname = outdbname;
        break;
    }

    /* open berbeley-db */
    if (bdbname) {
        if (db_create (&bdbp, NULL, 0)) {
            VERBOSE (ERROR, printf ("Error initializing database structure\n"));
        } else if (bdbp->open (bdbp, NULL, bdbname, NULL, DB_HASH, DB_CREATE, S_IRUSR | S_IWUSR)) {
            VERBOSE (ERROR, printf ("Error opening database '%s'\n", bdbname));
            return 0;
        }
    }

    /* open level-db */
    if (ldbname) {
        leveldb_options_t *options = leveldb_options_create ();
        leveldb_options_set_create_if_missing (options, 1);
        ldbp = leveldb_open (options, ldbname, &errptr);
        leveldb_options_destroy (options);
        if (!ldbp) {
            VERBOSE (ERROR, printf ("Error opening database '%s (%s)'\n", ldbname, errptr));
            return 0;
        }
        leveldb_free (errptr); errptr = NULL;
    }

    /* open ma-db */
    if (mdbname) {
        mdbp = madb_open_db (mdbname);
        if (!mdbp) {
            VERBOSE (ERROR, printf ("Error opening database '%s'\n", mdbname));
            return 0;
        }
    }

    return 1;
}

/* convertion functions */

int convert_bdb2ldb (char *indbname, char *outdbname)
{
    VERBOSE (DEBUG, printf ("Convert Berkeley-DB '%s' to Level-DB '%s'\n", indbname, outdbname));

    /* get all field from bdb and put them into ldb */
    leveldb_writeoptions_t *woptions = leveldb_writeoptions_create ();
    DBT key;
    memset (&key, 0, sizeof (DBT));
    key.flags = DB_DBT_MALLOC;
    DBT item;
    item.flags = DB_DBT_MALLOC;
    DBC *cursorp;
    bdbp->cursor (bdbp, NULL, &cursorp, 0);
    int nb_read = 0, nb_write = 0, ret;
    while (!(ret = cursorp->get (cursorp, &key, &item, DB_NEXT))) {
        nb_read++;

        VERBOSE (DEBUG, printf ("Inserting key '%s' (%d/%d) into db\n", (char *)key.data, key.size, item.size));
        leveldb_put (ldbp, woptions, (char *)key.data, key.size, (char *)item.data, item.size, &errptr);
        if (errptr) {
            VERBOSE (WARNING, printf ("Failed inserting into database (%s)\n", errptr));
        } else {
            nb_write++;
        }
        if (key.data) free (key.data); key.data = NULL;
        if (item.data) free (item.data); item.data = NULL;
        leveldb_free (errptr); errptr = NULL;

    }
    leveldb_writeoptions_destroy (woptions);
    VERBOSE (DEBUG, printf ("Read %d from bdb, write %d into ldb\n", nb_read, nb_write));

    return nb_read == nb_write;
}

int convert_bdb2mdb (char *indbname, char *outdbname)
{
    VERBOSE (DEBUG, printf ("Convert Berkeley-DB '%s' to Ma-DB '%s'\n", indbname, outdbname));

    /* get all field from bdb and put them into mdb */
    DBT key;
    memset (&key, 0, sizeof (DBT));
    key.flags = DB_DBT_MALLOC;
    DBT item;
    item.flags = DB_DBT_MALLOC;
    DBC *cursorp;
    bdbp->cursor (bdbp, NULL, &cursorp, 0);
    int nb_read = 0, nb_write = 0, ret;
    while (!(ret = cursorp->get (cursorp, &key, &item, DB_NEXT))) {
        nb_read++;

        VERBOSE (DEBUG, printf ("Inserting key '%s' (%d/%d) into db\n", (char *)key.data, key.size, item.size));
        if (!madb_update_item (mdbp, (char *)key.data, key.size, (char *)item.data, item.size)) {
            VERBOSE (WARNING, printf ("Failed inserting into database\n"));
        } else {
            nb_write++;
        }
        if (key.data) free (key.data); key.data = NULL;
        if (item.data) free (item.data); item.data = NULL;

    }
    VERBOSE (DEBUG, printf ("Read %d from bdb, write %d into mdb\n", nb_read, nb_write));

    return nb_read == nb_write;
}

int convert_ldb2bdb (char *indbname, char *outdbname)
{
    VERBOSE (DEBUG, printf ("Convert Level-DB '%s' to Berkeley-DB '%s'\n", indbname, outdbname));

    /* get all field from ldb and put them into bdb */
    int nb_read = 0, nb_write = 0;
    leveldb_readoptions_t *roptions = leveldb_readoptions_create ();
    leveldb_iterator_t *it = leveldb_create_iterator (ldbp, roptions);
    for (leveldb_iter_seek_to_first (it); leveldb_iter_valid (it); leveldb_iter_next (it)) {
        DBT key;
        memset (&key, 0, sizeof (key));
        size_t key_len;
        key.data = (char *) leveldb_iter_key (it, &key_len);
        key.size = key_len;
        DBT item;
        memset (&item, 0, sizeof (item));
        size_t item_len;
        item.data = (char *) leveldb_iter_value (it, &item_len);
        item.size = item_len;

        VERBOSE (DEBUG, printf ("Inserting key '%s' (%u/%u) into db\n", (char *)key.data, key.size, item.size));
        if (bdbp->put (bdbp, NULL, &key, &item, DB_OVERWRITE_DUP)) {
            VERBOSE (WARNING, printf ("Failed inserting into database\n"));
        } else {
            nb_write++;
        }
    }
    leveldb_iter_destroy (it);
    leveldb_readoptions_destroy (roptions);
    VERBOSE (DEBUG, printf ("Read %d from ldb, write %d into bdb\n", nb_read, nb_write));

    return nb_read == nb_write;
}

int convert_ldb2mdb (char *indbname, char *outdbname)
{
    VERBOSE (DEBUG, printf ("Convert Level-DB '%s' to Ma-DB '%s'\n", indbname, outdbname));

    /* get all field from ldb and put them into mdb */
    int nb_read = 0, nb_write = 0;
    leveldb_readoptions_t *roptions = leveldb_readoptions_create ();
    leveldb_iterator_t *it = leveldb_create_iterator (ldbp, roptions);
    for (leveldb_iter_seek_to_first (it); leveldb_iter_valid (it); leveldb_iter_next (it)) {
        size_t key_len = 0;
        char *key = (char *) leveldb_iter_key (it, &key_len);
        size_t item_len = 0;
        char *item = (char *) leveldb_iter_value (it, &item_len);

        VERBOSE (DEBUG, printf ("Inserting key '%s' (%"PRIuPTR"/%"PRIuPTR") into db\n", key, key_len, item_len));
        if (!madb_update_item (mdbp, key, key_len, item, item_len)) {
            VERBOSE (WARNING, printf ("Failed inserting into database\n"));
        } else {
            nb_write++;
        }
    }
    leveldb_iter_destroy (it);
    leveldb_readoptions_destroy (roptions);
    VERBOSE (DEBUG, printf ("Read %d from ldb, write %d into mdb\n", nb_read, nb_write));

    return nb_read == nb_write;
}

int convert_mdb2bdb (char *indbname, char *outdbname)
{
    VERBOSE (DEBUG, printf ("Convert Ma-DB '%s' to Berkeley-DB '%s'\n", indbname, outdbname));

    /* get all field from mdb and put them into bdb */
    madb_iterator_t *it = madb_new_iterator (mdbp);
    int nb_read = 0, nb_write = 0;
    for (madb_iterator_goto_begin (it); madb_iterator_valid (it); madb_iterator_next (it)) {
        nb_read++;
        DBT key;
        memset (&key, 0, sizeof (key));
        size_t key_len;
        key.data = madb_iterator_key (it, &key_len);
        key.size = key_len;
        DBT item;
        memset (&item, 0, sizeof (item));
        size_t item_len;
        item.data = madb_iterator_item (it, &item_len);
        item.size = item_len;

        VERBOSE (DEBUG, printf ("Inserting key '%s' (%u/%u) into db\n", (char *)key.data, key.size, item.size));
        if (bdbp->put (bdbp, NULL, &key, &item, DB_OVERWRITE_DUP)) {
            VERBOSE (WARNING, printf ("Failed inserting into database\n"));
        } else {
            nb_write++;
        }
        if (key.data) free (key.data);
        if (item.data) free (item.data);

    }
    madb_delete_iterator (it);
    VERBOSE (DEBUG, printf ("Read %d from mdb, write %d into bdb\n", nb_read, nb_write));

    return nb_read == nb_write;
}

int convert_mdb2ldb (char *indbname, char *outdbname)
{
    VERBOSE (DEBUG, printf ("Convert Ma-DB '%s' to Level-DB '%s'\n", indbname, outdbname));

    /* get all field from mdb and put them into ldb */
    leveldb_writeoptions_t *woptions = leveldb_writeoptions_create();
    madb_iterator_t *it = madb_new_iterator (mdbp);
    int nb_read = 0, nb_write = 0;
    for (madb_iterator_goto_begin (it); madb_iterator_valid (it); madb_iterator_next (it)) {
        nb_read++;

        size_t key_len = 0;
        char *key = madb_iterator_key (it, &key_len);
        size_t item_len = 0;
        char *item = madb_iterator_item (it, &item_len);
        VERBOSE (DEBUG, printf ("Inserting key '%s' (%"PRIuPTR"/%"PRIuPTR") into db\n", key, key_len, item_len));
        leveldb_put (ldbp, woptions, key, key_len, item, item_len, &errptr);
        if (errptr) {
            VERBOSE (WARNING, printf ("Failed inserting into database (%s)\n", errptr));
        } else {
            nb_write++;
        }
        if (key) free (key);
        if (item) free (item);
        leveldb_free (errptr); errptr = NULL;
    }
    madb_delete_iterator (it);
    leveldb_writeoptions_destroy (woptions);
    VERBOSE (DEBUG, printf ("Read %d from mdb, write %d into ldb\n", nb_read, nb_write));

    return nb_read == nb_write;
}

/* database dump function */

int dump_bdb (char *indbname)
{
    VERBOSE (DEBUG, printf ("Dump Berkeley-DB '%s'\n", indbname));

    DBT key;
    memset (&key, 0, sizeof (DBT));
    key.flags = DB_DBT_MALLOC;
    DBT item;
    item.flags = DB_DBT_MALLOC;
    DBC *cursorp;
    bdbp->cursor (bdbp, NULL, &cursorp, 0);
    int nb_read = 0, ret;
    size_t i;
    while (!(ret = cursorp->get (cursorp, &key, &item, DB_NEXT))) {
        nb_read++;

        printf ("key (%d): ", key.size);
        for (i = 0; i < key.size; i++)
            printf ("%02x", ((unsigned char *)key.data)[i]);
        printf (" ");
        printf ("item (%d): ", item.size);
        for (i = 0; i < item.size; i++)
            printf ("%02x", ((unsigned char *)item.data)[i]);
        printf ("\n");

        if (key.data) free (key.data); key.data = NULL;
        if (item.data) free (item.data); item.data = NULL;
    }

    return nb_read != 0;
}

int dump_ldb (char *indbname)
{
    VERBOSE (DEBUG, printf ("Dump Level-DB '%s'\n", indbname));

    int nb_read = 0;
    size_t i;
    leveldb_readoptions_t *roptions = leveldb_readoptions_create ();
    leveldb_iterator_t *it = leveldb_create_iterator (ldbp, roptions);
    for (leveldb_iter_seek_to_first (it); leveldb_iter_valid (it); leveldb_iter_next (it)) {
        nb_read++;
        size_t key_len = 0;
        unsigned char *key = (unsigned char *) leveldb_iter_key (it, &key_len);
        size_t item_len = 0;
        unsigned char *item = (unsigned char *) leveldb_iter_value (it, &item_len);

        printf ("key (%"PRIuPTR"): ", key_len);
        for (i = 0; i < key_len; i++)
            printf ("%02x", key[i]);
        printf (" ");
        printf ("item (%"PRIuPTR"): ", item_len);
        for (i = 0; i < item_len; i++)
            printf ("%02x", item[i]);
        printf ("\n");
    }
    leveldb_iter_destroy (it);
    leveldb_readoptions_destroy (roptions);

    return nb_read != 0;
}

int dump_mdb (char *indbname)
{
    VERBOSE (DEBUG, printf ("Dump Ma-DB '%s'\n", indbname));

    int nb_read = 0;
    size_t i;
    madb_iterator_t *it = madb_new_iterator (mdbp);
    for (madb_iterator_goto_begin (it); madb_iterator_valid (it); madb_iterator_next (it)) {
        nb_read++;

        size_t key_len = 0;
        unsigned char *key = (unsigned char *) madb_iterator_key (it, &key_len);
        size_t item_len = 0;
        unsigned char *item = (unsigned char *) madb_iterator_item (it, &item_len);

        printf ("key (%"PRIuPTR"): ", key_len);
        for (i = 0; i < key_len; i++)
            printf ("%02x", key[i]);
        printf (" ");
        printf ("item (%"PRIuPTR"): ", item_len);
        for (i = 0; i < item_len; i++)
            printf ("%02x", item[i]);
        printf ("\n");

        if (key) free (key);
        if (item) free (item);
    }
    madb_delete_iterator (it);

    return nb_read != 0;
}

/* close database */

int close_dbs ()
{
    int ret = 0;

    if (bdbp) {
        ret = bdbp->close (bdbp, 0);
        bdbp = NULL;
    }

    if (ldbp) {
        leveldb_close (ldbp);
        ret = 1;
        ldbp = NULL;

        leveldb_free (errptr);
        errptr = NULL;
    }

    if (mdbp) {
        ret = madb_close_db (mdbp);
        mdbp = NULL;
    }

    return ret;
}

/* help function */

void usage (int ret)
{
    FILE *fd = ret ? stderr : stdout;
    fprintf (fd, "usage: %s [-h] [-i db] [-o db] [-V] [-v int]\n", progname);
    fprintf (fd, "conversion tools (Berkeley-DB, Level-DB, Ma-DB)\n");
    fprintf (fd, " -d dump database: %d [none], %d [d], %d [l], %d [m] (%d)\n",
             NONE, BDB, LDB, MDB, IS_DUMP (way) ? way : NONE);
    fprintf (fd, " -h show help message\n");
    fprintf (fd, " -i define input db name\n");
    fprintf (fd, " -o define output db name\n");
    fprintf (fd, " -V show version\n");
    fprintf (fd, " -v verbose level (%d)\n", verbose);
    fprintf (fd, " -w conversion way: %d [b->l], %d [b->m], %d [l->b], %d [l->m], %d [m->b], %d [m->l] (%d)\n",
             BDB2LDB, BDB2MDB, LDB2BDB, LDB2MDB, MDB2BDB, MDB2LDB, IS_DUMP (way) ? NONE : way);

    exit (ret);
}

/* main */

int main (int argc, char *argv[])
{
    char indbname[1024] = {0};
    char outdbname[1024] = {0};

    /* basename of program file */
    progname = argv[0];
    int i = 0;
    while (progname[i] != '\0') {
        if (progname[i++] == '/') {
            progname += i;
            i = 0;
        }
    }

    int c;
    while ((c = getopt(argc, argv, "d:hi:o:Vv:w:")) != EOF) {
        switch (c) {
        case 'd':
            way = atoi (optarg);
            break;
        case 'h':
        default:
            usage (c != 'h');
            break;
        case 'i':
            strncpy (indbname, optarg, sizeof (indbname) - 1);
            break;
        case 'o':
            strncpy (outdbname, optarg, sizeof (outdbname) - 1);
            break;
        case 'V':
            printf ("Version: %s%s" HGID "%s\n", VERSION, strlen (HGID) ? " (" : "", strlen (HGID) ? ")" : "");
            return 0;
            break;
        case 'v':
            verbose = atoi (optarg);
            break;
        case 'w':
            way = atoi (optarg);
            break;
        }
    }
    if (argc - optind != 0) {
        fprintf (stderr, "%s: invalid option -- %s\n", progname, argv[optind]);
        usage (1);
    }
    if (strlen (indbname) == 0) {
        VERBOSE (ERROR, printf ("No input DB defined\n"));
        return 1;
    } else {
        struct stat buffer;
        if (stat (indbname, &buffer)) {
            VERBOSE (ERROR, printf ("Can't stat on DB '%s'\n", indbname));
            return 1;
        }
    }
    if (IS_DUMP (way)) {
        outdbname[0] = 0;
    } else if (strlen (outdbname) == 0) {
        VERBOSE (ERROR, printf ("No output DB defined\n"));
        return 1;
    }

    int ret = open_dbs (indbname, outdbname, way);

    if (!ret) {
        close_dbs ();
        return 1;
    }

    switch (way) {

    case BDB: /* dump Berkeley-DB */
        ret = dump_bdb (indbname);
        break;

    case LDB: /* dump Level-DB */
        ret = dump_ldb (indbname);
        break;

    case MDB: /* dump Ma-DB */
        ret = dump_mdb (indbname);
        break;

    case BDB2LDB: /* convert from Berkeley-DB to Level-DB */
        ret = convert_bdb2ldb (indbname, outdbname);
        break;

    case BDB2MDB: /* convert from Berkeley-DB to Ma-DB */
        ret = convert_bdb2mdb (indbname, outdbname);
        break;

    case LDB2BDB: /* convert from Level-DB to Berkeley-DB */
        ret = convert_ldb2bdb (indbname, outdbname);
        break;

    case LDB2MDB: /* convert from Level-DB to Ma-DB */
        ret = convert_ldb2mdb (indbname, outdbname);
        break;

    case MDB2BDB: /* convert from Ma-DB to Berkeley-DB */
        ret = convert_mdb2bdb (indbname, outdbname);
        break;

    case MDB2LDB: /* convert from Ma-DB to Level-DB */
        ret = convert_mdb2ldb (indbname, outdbname);
        break;
    }

    ret |= close_dbs ();

    return ret == 0;
}

/* test: rm -rf tests/db-[123].* */
/* test: {}; test $? -ne 0 */
/* test: {} -h */
/* test: {} -V */
/* test: {} -x; test $? -ne 0 */
/* test: {} x; test $? -ne 0 */
/* test: {} -w 5 -i tests/db.b -o tests/db-1.l; test $? -ne 0 */
/* test: {} -w 9 -i tests/db.b -o tests/db-1.l */
/* test: {} -w 13 -i tests/db.b -o tests/db-1.m */
/* test: {} -w 6 -i tests/db-1.l -o tests/db-2.b -v 0 */
/* test: {} -w 14 -i tests/db-1.l -o tests/db-2.m */
/* test: {} -w 7 -i tests/db-1.m -o tests/db-3.b -v 1 */
/* test: {} -w 11 -i tests/db-1.m -o tests/db-3.l -v 2 */
/* test: {} -d 1 -i tests/db.b | sort > tests/db-1.b.dump */
/* test: {} -d 2 -i tests/db-1.l | sort > tests/db-1.l.dump */
/* test: cmp tests/db-1.b.dump tests/db-1.l.dump */
/* test: VERBOSE_MADB_=0 {} -d 3 -i tests/db-1.m | sort > tests/db-1.m.dump */
/* test: cmp tests/db-1.b.dump tests/db-1.m.dump */
/* test: {} -d 1 -i tests/db-2.b | sort > tests/db-2.b.dump */
/* test: cmp tests/db-1.b.dump tests/db-2.b.dump */
/* test: VERBOSE_MADB_=0 {} -d 3 -i tests/db-2.m | sort > tests/db-2.m.dump */
/* test: cmp tests/db-1.b.dump tests/db-2.m.dump */
/* test: {} -d 1 -i tests/db-3.b | sort > tests/db-3.b.dump */
/* test: cmp tests/db-1.b.dump tests/db-3.b.dump */
/* test: {} -d 2 -i tests/db-3.l | sort > tests/db-3.l.dump */
/* test: cmp tests/db-1.b.dump tests/db-3.l.dump */
/* test: {} -d 1; test $? -ne 0 */
/* test: {} -d 1 -i /db.b; test $? -ne 0 */
/* test: {} -d 2 -i /db.l; test $? -ne 0 */
/* test: {} -d 3 -i /db.m; test $? -ne 0 */
/* test: {} -w 9 -i tests/db.b; test $? -ne 0 */
/* test: {} -w 9 -i tests/db.b -o /db.l; test $? -ne 0 */
/* test: {} -w 13 -i tests/db.b -o /db.m; test $? -ne 0 */
/* test: {} -w 6 -i tests/db-1.l -o /db.b; test $? -ne 0 */
