/*
  pedometer - a tools to retreive data and manage settings from Crivit Z31192 pedometer

  Copyright (c) 2014, Laurent Mazet <mazet@softndesign.org>
*/

#define __BDB_C__

#include <inttypes.h>
#include <malloc.h>
#include <leveldb/c.h>
#include <sys/stat.h>
#include <string.h>

#include "pedometer.h"

#include "db.h"

static char *dbfilename = NULL;

static leveldb_t *dbp = NULL;

void set_dbfilename (char *name)
{
    dbfilename  = name;
}

/* open db */

int open_db (void)
{
    static int init = 1;

    if (init) {
        leveldb_options_t *options = leveldb_options_create ();
        leveldb_options_set_create_if_missing (options, 1);
        char *errptr = NULL;
        dbp = leveldb_open (options, dbfilename, &errptr);
        if (!dbp) {
            VERBOSE (ERROR, printf ("Error opening database '%s (%s)'\n", dbfilename, errptr));
        }
        leveldb_options_destroy (options);
        leveldb_free (errptr);
        init = 0;
    }
    return (dbp != NULL);
}

/* close db */

void close_db ()
{
    if (dbp)
        leveldb_close (dbp);
    dbp = NULL;
}

/* fill db */

int fill_db (data_t *d)
{
    char *errptr = NULL;
    leveldb_writeoptions_t *options =leveldb_writeoptions_create();

    if (!open_db ())
        return 0;

    VERBOSE (DEBUG, printf ("Inserting key '%s' into db\n", d->date));
    leveldb_put (dbp, options, d->date, strlen (d->date) + 1, (char *)d, sizeof (data_t), &errptr);
    if (errptr) {
        VERBOSE (WARNING, printf ("Failed inserting into database (%s)\n", errptr));
    }
    leveldb_writeoptions_destroy (options);
    leveldb_free (errptr);

    return (errptr == NULL);
}

/* update db config */

int update_db_config (config_t *c)
{
    char *key = "config";
    char *errptr = NULL;
    leveldb_writeoptions_t *options = leveldb_writeoptions_create ();

    if (!open_db ())
        return 0;

    VERBOSE (DEBUG, printf ("Inserting config into db\n"));
    leveldb_put (dbp, options, key, strlen (key) + 1, (char *)c, sizeof (config_t), &errptr);
    if (errptr) {
        VERBOSE (WARNING, printf ("Failed inserting config into database (%s)\n", errptr));
    }
    leveldb_writeoptions_destroy (options);
    leveldb_free (errptr);

    return (errptr == NULL);
}

/* convert db */

int convert_db (config_t *new_config)
{
    config_t _db_config;
    config_t *db_config = &_db_config;
    memset (db_config, 0, sizeof (config_t));
    char *errptr = NULL;
    leveldb_readoptions_t *options = leveldb_readoptions_create ();

    if (!open_db ())
        return 0;

    /* look for db configuration */
    char *key = "config";
    size_t data_len = 0;
    char *data = leveldb_get (dbp, options, key, strlen (key) + 1, &data_len, &errptr);
    if (!data) {
        VERBOSE (ERROR, printf ("Config not found in database (%s)\n", errptr));
    } else {
        if (data_len > sizeof (config_t)) {
            VERBOSE (ERROR, printf ("Incorrect config size (%"PRIuPTR")\n", data_len));
        } else {
            memcpy (db_config, data, data_len);
        }
    }
    leveldb_free (data);
    leveldb_free (errptr);
    int ret = (errptr == NULL);
    
    /* check db unit system */
    if (db_config->system == system_undef_e) {
        printf ("what is the database system (1 for metric, 2 for imperial)? ");
        if ((scanf ("%u", &db_config->system)) && (db_config->system >= system_unknown_e)) {
            VERBOSE (WARNING, printf ("Unknown system (%u)\n", db_config->system));
            db_config->system = system_undef_e;
        }
        if (db_config->system != system_undef_e) {
            if (update_db_config (db_config)) {
                VERBOSE (DEBUG, printf ("Insert initial db config\n"));
            }
        }
    }
    VERBOSE (DEBUG, printf ("Config: %u -> %u\n", db_config->system, new_config->system));

    if (new_config->system) {
        if (new_config->system != db_config->system) {

            /* update db config */
            VERBOSE (DEBUG, printf ("Update database configuration\n"));
            if (update_db_config (new_config)) {

                /* convert db */
                if (db_config->system && new_config->system) {
                    VERBOSE (DEBUG, printf ("Convert database from %s to %s\n", (db_config->system == 1) ? "metric" : "imperial", (new_config->system == 1) ? "metric" : "imperial"));

                    leveldb_iterator_t *it = leveldb_create_iterator (dbp, options);
                    int nb = 0;
                    for (leveldb_iter_seek_to_first (it); leveldb_iter_valid (it); leveldb_iter_next (it)) {
                        size_t key_len = 0;
                        const char *key = leveldb_iter_key (it, &key_len);
                        size_t data_len = 0;
                        data_t *data = (data_t *) leveldb_iter_value (it, &data_len);
                        if ((key_len == 0) || (!key) || (key[key_len - 1] != 0) || (!data) ||
                            ((data_len != sizeof (config_t)) && (data_len != sizeof (data_t)))) {
                            VERBOSE (ERROR, printf ("Incorrect data read from database\n"));
                            ret = 0;
                        } else {
                            if (strncmp (key, "config", 6) != 0) {
                                data->distance = convert_gait (new_config->system, db_config->system, data->distance * 10);
                                data->distance = DIVROUND (data->distance, 1000);
                                fill_db (data);
                                nb++;
                            }
                        }
                    }
                    leveldb_iter_destroy (it);
                    VERBOSE (DEBUG, printf ("Process %d records\n", nb));
                }
            }

        } else {
            VERBOSE (WARNING, printf ("Nothing to do\n"));
        }

    } else {
        VERBOSE (WARNING, printf ("Incoherent unit systeme (db or request), nothing to do\n"));
    }
    leveldb_readoptions_destroy (options);

    return ret;
}

/* extract db */

int extract_db (data_t **table)
{
    data_t _data;
    data_t *data = &_data;
    memset (data, 0, sizeof (data_t));
    leveldb_readoptions_t *options = leveldb_readoptions_create ();

    if (!open_db ())
        return 0;

    leveldb_iterator_t *it = leveldb_create_iterator (dbp, options);
    int i, nb = 0;
    for (leveldb_iter_seek_to_first (it); leveldb_iter_valid (it); leveldb_iter_next (it)) {
        size_t key_len = 0;
        const char *key = leveldb_iter_key (it, &key_len);
        size_t data_len = 0;
        data_t *data = (data_t *) leveldb_iter_value (it, &data_len);
        if ((key_len == 0) || (!key) || (key[key_len - 1] != 0) || (!data) ||
            ((data_len != sizeof (config_t)) && (data_len != sizeof (data_t)))) {
            VERBOSE (ERROR, printf ("Incorrect data read from database\n"));
        } else {
            if (strncmp (key, "config", 6) != 0) {
                *table = (data_t *) realloc (*table, (nb + 2) * sizeof (data_t));
                for (i = 0; i < nb; i++)
                    if (strcmp (data->date, (*table + i)->date) > 0)
                        break;
                if (i < nb)
                    memmove (*table + i + 1, *table + i, (nb - i) * sizeof (data_t));
                memcpy (*table + i, data, sizeof (data_t));
                nb++;
            }
        }
    }
    leveldb_iter_destroy (it);
    leveldb_readoptions_destroy (options);

    return nb;
}
