/*
  pedometer - a tools to retreive data and manage settings from Crivit Z31192 pedometer

  Copyright (c) 2014, Laurent Mazet <mazet@softndesign.org>
*/

#define __BDB_C__

#include <inttypes.h>
#include <malloc.h>
#include <snd/madb.h>
#include <sys/stat.h>
#include <string.h>

#include "pedometer.h"

#include "db.h"

static char *dbfilename = NULL;

static madb_t *dbp = NULL;

void set_dbfilename (char *name)
{
    dbfilename  = name;
}

/* open db */

int open_db (void)
{
    static int init = 1;

    if (init) {
        dbp = madb_open_db (dbfilename);
        if (!dbp) {
            VERBOSE (ERROR, printf ("Error opening database '%s)'\n", dbfilename));
        }
        init = 0;
    }
    return (dbp != NULL);
}

/* close db */

void close_db ()
{
    if (dbp)
        madb_close_db (dbp);
    dbp = NULL;
}

/* fill db */

int fill_db (data_t *d)
{
    if (!open_db ())
        return 0;

    VERBOSE (DEBUG, printf ("Inserting key '%s' into db\n", d->date));
    int rc = madb_update_item (dbp, d->date, strlen (d->date) + 1, (char *)d, sizeof (data_t));
    if (rc == 0) {
        VERBOSE (WARNING, printf ("Failed inserting into database\n"));
    }

    return (rc != 0);
}

/* update db config */

int update_db_config (config_t *c)
{
    char *key = "config";

    if (!open_db ())
        return 0;

    VERBOSE (DEBUG, printf ("Inserting config into db\n"));
    int rc = madb_update_item (dbp, key, strlen (key) + 1, (char *)c, sizeof (config_t));
    if (rc == 0) {
        VERBOSE (WARNING, printf ("Failed inserting config into database\n"));
    }

    return (rc != 0);
}

/* convert db */

int convert_db (config_t *new_config)
{
    config_t _db_config;
    config_t *db_config = &_db_config;
    memset (db_config, 0, sizeof (config_t));

    if (!open_db ())
        return 0;

    /* look for db configuration */
    char *key = "config";
    size_t data_len = 0;
    char *data = madb_get_item (dbp, key, strlen (key) + 1, &data_len);
    if (!data) {
        VERBOSE (ERROR, printf ("Config not found in database\n"));
    } else {
        if (data_len > sizeof (config_t)) {
            VERBOSE (ERROR, printf ("Incorrect config size (%"PRIuPTR")\n", data_len));
        } else {
            memcpy (db_config, data, data_len);
        }
    }
    free (data);
    
    /* check db unit system */
    if (db_config->system == system_undef_e) {
        printf ("what is the database system (1 for metric, 2 for imperial)? ");
        if ((scanf ("%u", &db_config->system)) && (db_config->system >= system_unknown_e)) {
            VERBOSE (WARNING, printf ("Unknown system (%u)\n", db_config->system));
            db_config->system = system_undef_e;
        }
        if (db_config->system != system_undef_e) {
            if (update_db_config (db_config)) {
                VERBOSE (DEBUG, printf ("Insert initial db config\n"));
            }
        }
    }
    VERBOSE (DEBUG, printf ("Config: %u -> %u\n", db_config->system, new_config->system));

    if (new_config->system) {
        if (new_config->system != db_config->system) {

            /* update db config */
            VERBOSE (DEBUG, printf ("Update database configuration\n"));
            if (update_db_config (new_config)) {

                /* convert db */
                if (db_config->system && new_config->system) {
                    VERBOSE (DEBUG, printf ("Convert database from %s to %s\n", (db_config->system == 1) ? "metric" : "imperial", (new_config->system == 1) ? "metric" : "imperial"));

                    madb_iterator_t *it = madb_new_iterator (dbp);
                    int nb = 0;
                    for (madb_iterator_goto_begin (it); madb_iterator_valid (it); madb_iterator_next (it)) {
                        size_t key_len = 0;
                        char *key = madb_iterator_key (it, &key_len);
                        size_t data_len = 0;
                        data_t *data = (data_t *) madb_iterator_item (it, &data_len);
                        if ((key_len == 0) || (!key) || (key[key_len - 1] != 0) || (!data) ||
                            ((data_len != sizeof (config_t)) && (data_len != sizeof (data_t)))) {
                            VERBOSE (ERROR, printf ("Incorrect data read from database\n"));
                        } else {
                            if (strncmp (key, "config", 6) != 0) {
                                data->distance = convert_gait (new_config->system, db_config->system, data->distance * 10);
                                data->distance = DIVROUND (data->distance, 1000);
                                fill_db (data);
                                nb++;
                            }
                        }
                        if (key)
                            free (key);
                        if (data)
                            free (data);
                    }
                    madb_delete_iterator (it);
                    VERBOSE (DEBUG, printf ("Process %d records\n", nb));
                }
            }

        } else {
            VERBOSE (WARNING, printf ("Nothing to do\n"));
        }

    } else {
        VERBOSE (WARNING, printf ("Incoherent unit systeme (db or request), nothing to do\n"));
    }

    return 1;
}

/* extract db */

int extract_db (data_t **table)
{
    data_t _data;
    data_t *data = &_data;
    memset (data, 0, sizeof (data_t));

    if (!open_db ())
        return 0;

    madb_iterator_t *it = madb_new_iterator (dbp);
    int i, nb = 0;
    for (madb_iterator_goto_begin (it); madb_iterator_valid (it); madb_iterator_next (it)) {
        size_t key_len = 0;
        char *key = madb_iterator_key (it, &key_len);
        size_t data_len = 0;
        data_t *data = (data_t *) madb_iterator_item (it, &data_len);
        if ((key_len == 0) || (!key) || (key[key_len - 1] != 0) || (!data) ||
            ((data_len != sizeof (config_t)) && (data_len != sizeof (data_t)))) {
            VERBOSE (ERROR, printf ("Incorrect data read from database\n"));
        } else {
            if (strncmp (key, "config", 6) != 0) {
                *table = (data_t *) realloc (*table, (nb + 2) * sizeof (data_t));
                for (i = 0; i < nb; i++)
                    if (strcmp (data->date, (*table + i)->date) > 0)
                        break;
                if (i < nb)
                    memmove (*table + i + 1, *table + i, (nb - i) * sizeof (data_t));
                memcpy (*table + i, data, sizeof (data_t));
                nb++;
            }
        }
        if (key)
            free (key);
        if (data)
            free (data);
    }
    madb_delete_iterator (it);

    return nb;
}
